(* File: mizartypechecker.ml *)
(* Author: Chad E Brown *)
(* Modified copy of satallax.ml, June 2017 *)

open Mizartypecheckermain
open Syntax
open Mizartypeinfo
open State
open Log
open Printf
open Error

let status_str s = "% SZS status " ^ s
let error_str = status_str "Error"

let error_reaction = function
| InputHelpError(x) -> 1, error_str :: sprintf "Input Error: %s" x :: help_lines
| InputError(x) -> 1, [error_str; sprintf "Input Error: %s" x]
| FileNotFound(x) -> 1, [error_str; sprintf "File Not Found: %s" x]
| Parsing.Parse_error -> 2, [error_str; "Syntax Error"]
| ParsingError(l1,i1,l2,i2,x) -> 2, [error_str; sprintf "Parsing Error: %s from line %d char %d to line %d char %d\n" x l1 i1 l2 i2]
| ExpectedTypeError(m,a,b) -> 2, [error_str; sprintf "%s\nhas type\n%s\nexpected type\n%s\n" (pretrm_str m) (stp_str b) (stp_str a)]
| GenericError(x) -> 2, [error_str; x]
| GenericSyntaxError(x) -> 2, [error_str; x]
| Redeclaration(x) -> 2, [error_str; sprintf "%s cannot be redeclared" x]
| IncompleteSatisfiable -> 6, []
| Timeout -> 5, if (!verbosity > 0) then [status_str ("Timeout")] else []
| e -> 3, if (!verbosity > 0) then [error_str; sprintf "Exception: %s" (Printexc.to_string e)] else []

let pn_head p =
  match neg_body p with
  | Some(n) ->
      begin
	match head_spine n with
	| (Name(h,_),_) -> (h,false)
	| _ -> raise Not_found
      end
  | _ ->
      match head_spine p with
      | (Name(h,_),_) -> (h,true)
      | _ -> raise Not_found

let pn_head_in p hl =
  let (h,pol) = pn_head p in
  if List.mem h hl then (h,pol) else raise Not_found

let printheads pl =
  List.iter (fun m -> try let (h,p) = pn_head m in Printf.printf "  %s %s" h (if p then "+" else "-") with Not_found -> ()) pl;
  Printf.printf "\n"

let grepconclheads f hl =
  let c = open_in_bin f in
  mizar_symbol_order_pred := input_value c;
  mizar_name_tp := input_value c;
  mizar_pattern_list := input_value c;
  try
    while true do
      let mta : mizartypingaxiom = input_value c in
      match mta with
      | MizarFunctionType(ax,fl,i,pl,bl) ->
	  List.iter (fun b -> try let (h,p) = pn_head_in b hl in Printf.printf "%s %s %s\n" ax h (if p then "+" else "-"); printheads pl with Not_found -> ()) bl
      | MizarTypeHierarchyStrict(ax,fl,i,(p::pl),bl) ->
	  let bl2 = List.filter (fun m -> not (List.mem m (p::pl))) bl in
	  List.iter (fun b -> try let (h,p) = pn_head_in b hl in Printf.printf "%s %s %s\n" ax h (if p then "+" else "-"); printheads pl with Not_found -> ()) bl2
      | MizarTypeHierarchy(ax,fl,i,(p::pl),bl) ->
	  let bl2 = List.filter (fun m -> not (List.mem m (p::pl))) bl in
	  List.iter (fun b -> try let (h,p) = pn_head_in b hl in Printf.printf "%s %s %s\n" ax h (if p then "+" else "-"); printheads pl with Not_found -> ()) bl2
      | MizarTypingGeneric(ax,i,pl,ml) ->
	  let mlsimp = List.filter (fun m -> not (List.mem m pl) && (let (h,s) = head_spine m in try ignore (List.find (fun n -> match n with DB(_,_) -> false | _ -> true) s); false with Not_found -> true)) ml in
	  if not (mlsimp = []) then
	    List.iter (fun b -> try let (h,p) = pn_head_in b hl in Printf.printf "%s %s %s\n" ax h (if p then "+" else "-"); printheads pl with Not_found -> ()) mlsimp
      | MizarRedefinition(ax,i,pl,f,s,d) -> if List.mem f hl then Printf.printf "redef %s %s\n" ax f
      | _ -> ()
    done
  with End_of_file ->
    close_in c
    
let greppremheads f hl =
  let c = open_in_bin f in
  mizar_symbol_order_pred := input_value c;
  mizar_name_tp := input_value c;
  mizar_pattern_list := input_value c;
  try
    while true do
      let mta : mizartypingaxiom = input_value c in
      match mta with
      | MizarFunctionType(ax,fl,i,pl,bl) ->
	  List.iter (fun b -> try let (h,p) = pn_head_in b hl in Printf.printf "%s %s %s\n" ax h (if p then "+" else "-"); printheads bl with Not_found -> ()) pl
      | MizarTypeHierarchyStrict(ax,fl,i,(p::pl),bl) ->
	  List.iter (fun b -> try let (h,p) = pn_head_in b hl in Printf.printf "%s %s %s\n" ax h (if p then "+" else "-"); printheads bl with Not_found -> ()) pl
      | MizarTypeHierarchy(ax,fl,i,(p::pl),bl) ->
	  List.iter (fun b -> try let (h,p) = pn_head_in b hl in Printf.printf "%s %s %s\n" ax h (if p then "+" else "-"); printheads bl with Not_found -> ()) pl
      | MizarTypingGeneric(ax,i,pl,ml) ->
	  List.iter (fun b -> try let (h,p) = pn_head_in b hl in Printf.printf "%s %s %s\n" ax h (if p then "+" else "-"); printheads ml with Not_found -> ()) pl
      | MizarRedefinition(ax,i,pl,f,s,d) ->
	  List.iter (fun b -> try let (h,p) = pn_head_in b hl in Printf.printf "%s %s %s\n" ax h (if p then "+" else "-") with Not_found -> ()) pl
      | _ -> ()
    done
  with End_of_file ->
    close_in c
    
let raisebadcall() =
  raise (GenericError("should be mizartypechecker -create <createmizartypingaxiomsfile> <thffile> or mizartypechecker -model <existingmizartypingaxiomsfile> <createmodelsfile> or mizartypechecker -check <existingmizartypingaxiomsfile> <thffile> or mizartypechecker -satisfy <modelsfile> <thffile>"))

let number_p x =
  if String.length x = 0 then
    false
  else
    let c = Char.code x.[0] in
    c >= 48 && c < 58
      
let mizar_numbers_hack f i = (*** just add more types I think numbers should satisfy but are not currently derived; at this point they do seem to be derived, but it's still faster to just assume them ***)
  if i = 0 && number_p f then
    begin
      let io = Ar(Base("$i"),Prop) in
      let fi = Name(f,Base("$i")) in
      let q p = Ap(Name(p,io),fi) in
      let bl = [q "v7_ordinal1";q "v1_int_1";q "v3_ordinal1"] in
      Hashtbl.add mizar_func_types f ("numberhack",0,[],bl)
    end

let rec specialargs cla =
  match cla with
  | ("-maxelaborateiters"::x::clar) ->
      maxelaborateiters := int_of_string x;
      specialargs clar
  | ("-elaboratelim"::x::clar) ->
      elaboratelim := int_of_string x;
      specialargs clar
  | ("-elaboratetimeout"::x::clar) ->
      elaboratetimeout := float_of_string x;
      specialargs clar
  | ("-acceptcommonmizartypes"::clar) ->
      acceptcommonmizartypes := true;
      specialargs clar
  | _ -> cla
    
let _ =
  try
    let cla = ref [] in
    for i = (Array.length Sys.argv) - 1 downto 1 do
      cla := Sys.argv.(i) :: !cla
    done;
    cla := specialargs !cla;
    match !cla with
    | ["-create";mtaf;mttf] -> (*** mizartypechecker -create <mizartypingaxiomsfile> <allofmizarthffile> ***)
	collecttypingaxioms := true;
	mizartypechecker_main false mttf;
	let c = open_out_bin mtaf in
	output_value c mizar_symbol_order_pred;
	output_value c name_tp;
	output_value c !mizar_pattern_list;
	List.iter (fun m -> output_value c m) !typingaxiomslist;
	close_out c
    | ["-check";mtaf;qf] -> (*** mizartypechecker -check <mizartypingaxiomsfile> <thffilewithcheckelabqueries> ***)
	begin
	  let c = open_in_bin mtaf in
	  mizar_symbol_order_pred := input_value c;
	  mizar_name_tp := input_value c;
	  mizar_pattern_list := input_value c;
	  typingaxiomslist := [];
	  try
	    while true do
	      let mta : mizartypingaxiom = input_value c in
	      typingaxiomslist := mta::!typingaxiomslist;
	      match mta with
	      | MizarFunctionType(ax,fl,i,pl,bl) ->
		  List.iter (fun f -> Hashtbl.add mizar_func_types f (ax,i,pl,bl); mizar_numbers_hack f i) fl
	      | MizarTypeHierarchyStrict(ax,fl,i,(p::pl),bl) ->
		  let bl2 = List.filter (fun m -> not (List.mem m (p::pl))) bl in
		  if not (bl2 = []) then
		    begin
		      List.iter (fun f -> Hashtbl.add mizar_typing f (ax,i,p,pl,bl2)) fl
		    end
	      | MizarTypeHierarchy(ax,fl,i,(p::pl),bl) ->
		  let bl2 = List.filter (fun m -> not (List.mem m (p::pl))) bl in
		  if not (bl2 = []) then
		    begin
		      List.iter (fun f -> Hashtbl.add mizar_typing f (ax,i,p,pl,bl2)) fl
		    end
	      | MizarTypingGeneric(ax,i,pl,ml) ->
		  begin
		    let rec db_occur l i a m =
		      match m with
		      | DB(j,_) when j >= l ->
			  let k = j - l in
			  if k < i then
			    a.(k) <- true
			  else
			    raise Exit
		      | Ap(m1,m2) ->
			  db_occur l i a m1;
			  db_occur l i a m2
		      | Lam(_,m1) ->
			  db_occur (l+1) i a m1
		      | _ -> ()
		    in
		    let mlsimp = List.filter (fun m -> not (List.mem m pl) && (let (h,s) = head_spine m in try ignore (List.find (fun n -> match n with DB(_,_) -> false | _ -> true) s); false with Not_found -> true)) ml in
		    if not (mlsimp = []) then
		      begin
			let pl1 = ref pl in
			let pr1 = ref [] in
			while not (!pl1 = []) do
			  match !pl1 with
			  | [] -> () (*** impossible ***)
			  | (prim::plr) ->
			      pl1 := plr;
			      let alldboccur =
				begin
				  let a = Array.create i false in
				  try
				    db_occur 0 i a prim;
				    for j = 0 to i-1 do
				      if not a.(j) then raise Exit
				    done;
				    true
				  with Exit -> false
				end
			      in
			      if alldboccur then
				begin
				  match neg_body prim with
				  | Some(nprim) ->
				      begin
					let (h,s) = head_spine nprim in
					match h with
					| Name(f,_) -> Hashtbl.add mizar_typing_gen_forward f (ax,i,prim,(!pr1 @ !pl1),mlsimp)
					| _ -> ()
				      end
				  | None ->
				      begin
					let (h,s) = head_spine prim in
					match h with
					| Name(f,_) -> Hashtbl.add mizar_typing_gen_forward f (ax,i,prim,(!pr1 @ !pl1),mlsimp)
					| _ -> ()
				      end
				end;
			      pr1 := prim::!pr1
			done
		      end;
		    List.iter
		      (fun concl ->
			if not (List.mem concl pl) then
			  let alldboccur =
			    begin
			      let a = Array.create i false in
			      try
				db_occur 0 i a concl;
				for j = 0 to i-1 do
				  if not a.(j) then raise Exit
				done;
				true
			      with Exit -> false
			    end
			  in
			  if alldboccur then
			    begin
			      match neg_body concl with
			      | Some(nconcl) ->
				  begin
				    let (mh,ms) = head_spine nconcl in
				    match ms with
				    | [] -> ()
				    | (a::r) ->
					let (h,s) = head_spine a in
					match h with
					| Name(f,_) -> Hashtbl.add mizar_typing_gen_backward f (ax,i,a,pl,concl)
					| _ -> ()
				  end
			      | None ->
				  let (mh,ms) = head_spine concl in
				  match ms with
				  | [] -> ()
				  | (a::r) ->
				      let (h,s) = head_spine a in
				      match h with
				      | Name(f,_) -> Hashtbl.add mizar_typing_gen_backward f (ax,i,a,pl,concl)
				      | _ -> ()
			    end)
		      ml
		  end
	      | MizarRedefinition(ax,i,pl,f,s,d) ->
		  Hashtbl.add mizar_redef f (ax,i,pl,s,d)
	      | _ -> ()
	    done
	  with End_of_file ->
	    close_in c;
	    List.iter (fun (nname,i,pl,rtp,al,pd,d) -> Hashtbl.add mizar_pattern_table nname (i,pl,rtp,al,pd,d)) !mizar_pattern_list;
	    mizartypechecker_elaborate qf
	end
    | ["-printinfo";mtaf] -> (*** mizartypechecker -printinfo <mizartypingaxiomsfile> ***)
	begin
	  let maxquantdepth = ref 0 in
	  let mod2quants = ref 0 in
	  let mod4quants = ref 0 in
	  let mod8quants = ref 0 in
	  let mod16quants = ref 0 in
	  let quantdepth i =
	    maxquantdepth := max !maxquantdepth i;
	    mod2quants := !mod2quants + 1 lsl i;
	    mod4quants := !mod4quants + 1 lsl (2*i);
	    mod8quants := !mod8quants + 1 lsl (3*i);
	    mod16quants := !mod16quants + 1 lsl (4*i);
	  in
	  let c = open_in_bin mtaf in
	  mizar_symbol_order_pred := input_value c;
	  mizar_name_tp := input_value c;
	  let mizar_tp_name : (stp,string) Hashtbl.t = Hashtbl.create 1000 in
	  let tpl = ref [] in
	  Hashtbl.iter (fun x a -> if not (x.[0] = 'n') then (Hashtbl.add mizar_tp_name a x; if not (List.mem a !tpl) then tpl := a::!tpl)) !mizar_name_tp;
	  List.iter (fun a -> Printf.printf "%s : %d\n" (stp_str a) (List.length (Hashtbl.find_all mizar_tp_name a))) !tpl;
	  mizar_pattern_list := input_value c;
	  typingaxiomslist := [];
	  try
	    while true do
	      let mta : mizartypingaxiom = input_value c in
	      typingaxiomslist := mta::!typingaxiomslist;
	      match mta with
	      | MizarFunctionType(ax,fl,i,pl,bl) -> quantdepth i
	      | MizarTypeHierarchyStrict(ax,fl,i,(p::pl),bl) ->
		  let bl2 = List.filter (fun m -> not (List.mem m (p::pl))) bl in
		  if not (bl2 = []) then quantdepth i
	      | MizarTypeHierarchy(ax,fl,i,(p::pl),bl) ->
		  let bl2 = List.filter (fun m -> not (List.mem m (p::pl))) bl in
		  if not (bl2 = []) then quantdepth i
	      | MizarTypingGeneric(ax,i,pl,ml) -> quantdepth i
	      | MizarRedefinition(ax,i,pl,f,s,d) -> quantdepth i
	      | _ -> ()
	    done
	  with End_of_file ->
	    close_in c;
	    Printf.printf "maxquantdepth = %d\nmod2quants = %d\nmod4quants = %d\nmod8quants = %d\nmod16quants = %d\n" !maxquantdepth !mod2quants !mod4quants !mod8quants !mod16quants
	end
    | ("-grepconclheads"::mtaf::hl) ->
	grepconclheads mtaf hl
    | ("-greppremheads"::mtaf::hl) ->
	greppremheads mtaf hl
    | ["-printpatterns";mtaf] ->
	begin
	  let c = open_in_bin mtaf in
	  mizar_symbol_order_pred := input_value c;
	  mizar_name_tp := input_value c;
	  mizar_pattern_list := input_value c;
	  Printf.printf "%d patterns\n" (List.length !mizar_pattern_list);
	  List.iter
	    (fun (x,i,pl,_,al,pd,d) ->
	      Printf.printf "(\"%s\" %d (" x i;
	      List.iter (fun a -> Printf.printf "%d " a) al;
	      Printf.printf ") (";
	      let (h,s) = head_spine pd in
	      Printf.printf "\"%s\"" (trm_str h);
	      List.iter (fun a -> match a with DB(i,_) -> Printf.printf " %d" i | _ -> Printf.printf " NIL") s;
	      Printf.printf ") (";
	      let (h,s) = head_spine d in
	      Printf.printf "\"%s\"" (trm_str h);
	      List.iter (fun a -> match a with DB(i,_) -> Printf.printf " %d" i | _ -> Printf.printf " NIL") s;
	      Printf.printf "))\n")
	    !mizar_pattern_list;
	  typingaxiomslist := [];
	  close_in c;
	end
    | _ -> raisebadcall()
  with e ->
    let (code, msg) = error_reaction e in
    List.iter print_endline msg;
    exit code

