(* File: mizartypecheckermain.ml *)
(* ported from satallaxmain.ml Aug 2017 *)

open Treetype
open Syntax
open Mizartypeinfo
open State
open Log
open Error

let debug = false
let mizvarcnt = ref 0
let mizwidenlimit = 7

exception InputHelpError of string
exception InputError of string

type globalids_type =
    { seti : identifier;
      booli : identifier;
      typei : identifier;
      attypei : identifier;
      atlistofadjectives : identifier;
      adjective : identifier;
      attribute : identifier;
      numi : identifier;
      varlist : identifier;
      var : identifier;
      lparen : identifier;
      rparen : identifier;
      lparenhash : identifier;
      rparenhash : identifier;
      comma : identifier;
      ddcomma : identifier;
      fraenkel : identifier;
      fraenkelvars : identifier;
      ddverum : identifier;
      ddover : identifier;
      ddthe : identifier;
      ddfor : identifier;
      ddex : identifier;
      ddof : identifier;
      ddis : identifier;
      ddbe : identifier;
      ddbeing : identifier;
      ddst : identifier;
      ddholds : identifier;
      ddnon : identifier;
      ddnot : identifier;
      ddamp : identifier;
      ddor : identifier;
      ddimplies : identifier;
      ddiff : identifier;
      ddif : identifier;
      ddotherwise : identifier;
      ddsetenumlbrace : identifier;
      ddsetenumrbrace : identifier;
      ddfraenkellbrace : identifier;
      ddfraenkelrbrace : identifier;
      ddfraenkelcolon : identifier;
      ddfraenkelwhere : identifier;
    }

let globalids : globalids_type =
  { seti = identifier_of_string "Set";
    booli = identifier_of_string "Bool";
    typei = identifier_of_string "Type";
    attypei = identifier_of_string "@Type";
    atlistofadjectives = identifier_of_string "@ListOfAdjectives";
    adjective = identifier_of_string "Adjective";
    attribute = identifier_of_string "Attribute";
    numi = identifier_of_string "Num";
    varlist = identifier_of_string "Varlist";
    var = identifier_of_string "Var";
    lparen = identifier_of_string "(";
    rparen = identifier_of_string ")";
    lparenhash = identifier_of_string "(#";
    rparenhash = identifier_of_string "#)";
    comma = identifier_of_string ",";
    ddcomma = identifier_of_string "$$,";
    fraenkel = identifier_of_string "Fraenkel";
    fraenkelvars = identifier_of_string "FraenkelVars";
    ddverum = identifier_of_string "$$verum";
    ddover = identifier_of_string "$$over";
    ddthe = identifier_of_string "$$the";
    ddfor = identifier_of_string "$$for";
    ddex = identifier_of_string "$$ex";
    ddof = identifier_of_string "$$of";
    ddis = identifier_of_string "$$is";
    ddbe = identifier_of_string "$$be";
    ddbeing = identifier_of_string "$$being";
    ddst = identifier_of_string "$$st";
    ddholds = identifier_of_string "$$holds";
    ddnot = identifier_of_string "$$not";
    ddnon = identifier_of_string "$$non";
    ddamp = identifier_of_string "$$&";
    ddor = identifier_of_string "$$or";
    ddimplies = identifier_of_string "$$implies";
    ddiff = identifier_of_string "$$iff";
    ddif = identifier_of_string "$$if";
    ddotherwise = identifier_of_string "$$otherwise";
    ddsetenumlbrace = identifier_of_string "setenumlbrace";
    ddsetenumrbrace = identifier_of_string "setenumrbrace";
    ddfraenkellbrace = identifier_of_string "fraenkellbrace";
    ddfraenkelrbrace = identifier_of_string "fraenkelrbrace";
    ddfraenkelcolon = identifier_of_string "fraenkelcolon";
    ddfraenkelwhere = identifier_of_string "fraenkelwhere";
  }

let itp = Base("$i")
let setoftm = Lam(Ar(itp,Prop),Ap(Choice(itp),Lam(itp,Ap(Forall(itp),Lam(itp,Ap(Ap(Eq(Prop),Ap(Ap(Name("r2_hidden",Ar(itp,Ar(itp,Prop))),DB(0,itp)),DB(1,itp))),Ap(DB(2,Ar(itp,Prop)),DB(0,itp))))))));;

let coqfile = ref ""
let flag_overrides = ref []
let schedule_files : string list ref = ref []
let preparsed = ref false
let starttime = ref 0.0
let endtime = ref 0.0

let printnumaxiomsflag : bool ref = ref false
let selectaxiomsflag : bool ref = ref false

let fclassifier_file : string option ref = ref None
let tclassifier_file : string option ref = ref None
let simple_file : string option ref = ref None

let sine_tolerance_vary = ref true
let sine_generality_threshold_vary = ref true
let sine_tolerance = ref 1.2 (*** can be set by command line --sine_tolerance t where t should >= 1.0; if none is set then slice over different values. ***)
let sine_depth = ref None (*** can be set by command line --sine_depth d where d should be an >= 0.0; if none is set then slice over different depths (more or less) ***)
let sine_generality_threshold = ref 0 (*** can be set by command line --sine_generality_threshold g; where g should be an integer >= 0 ***)

let help_lines =
[ "Usage: satallax [-[VvNhi]] [-verb <int>] [-P <PicoMus>] [-M <modedir>] [-s <schedule>] [-m <mode>] [-flag <name> <value>] [-t <timeout in seconds>] [-inferences <int>] [-p <pfformat>] <problemfile>"
; "-M <dir> : Set the directory in which the mode/schedule files are stored."
; "       The default mode directory is the 'modes' subdirectory of the Satallax directory."
; "-s : Schedule of modes to try (previously called the 'strategy schedule')"
; "-m : Mode"
; "-P <file> : PicoMus executable file"
; "-E <file> : E prover executable file"
; "-V : Print version number and quit"
; "-v : Verbose"
; "-h : recognize hashroots for special treatment"
; "-i : given problem is incomplete version of the real problem, so do not report Satisfiable or CounterSatisfiable"
; "-N : Try to determine if the problem is a non-theorem (Satisfiable or CounterSatisfiable)"
; "-verb <int> : Verbosity of given level, -verb 0 means silent"
; "-c [<file>|-] : Create a Coq version of problem, with a proof script if -p is included and proof search succeeds"
; "-C : The problem is given as a Coq file instead of as a THF file."
; "-G : A Coq file containing multiple conjectures is given. Try to prove each of them independently."
; "-p [tstp|coqscript|coqspfterm|hocore|modeltrue|model|isar|info|useful|formdeps]: Output proof/proof object/proof info"
; "-pfusefulout <file>: File in which to save info about what was useful in the proof"
; "-pfformdepsout <file>: File in which to save info about what formulas and dependencies were in the proof"
; "--sine_tolerance <float>: Set the sine_tolerance (t value) to some value >= 1.0"
; "--sine_depth <float>: Set the sine_depth (d value) to some value >= 0.0"
; "--sine_generality_threshold <int>: Set the sine_generality_threshold (g value) to some value >= 0"
]

let read_proofkind = function
  "tstp" -> TSTP
| "coqscript" -> CoqScript
| "coqspfterm" -> CoqSPfTerm
| "hocore" -> HOCore
| "model" -> Model
| "modeltrue" -> ModelTrue
| "isar" -> IsarScript
| "info" -> PfInfo
| "useful" -> PfUseful
| "formdeps" -> PfFormdeps
| p -> raise (InputHelpError ("Unknown kind of proof " ^ p ^ " for -p"))

let slaveargs = ref []
    
let enslave args = slaveargs := List.rev_append args !slaveargs

let process_short_cmd_line_arg = function
  'v' -> enslave ["-v"]; verbosity := 5
| 'i' -> enslave ["-i"]; completep := false
|  a  -> raise (InputHelpError ("Unknown command line argument " ^ String.make 1 a))

let set_problemfile p =
  if !problemfile = "" then problemfile := p
  else raise (InputHelpError ("Multiple problem files passed: " ^ !problemfile
  ^ " and " ^ p))

let code_status = function
  | (true , Some _) -> 10, "CounterSatisfiable"
  | (true , None  ) -> 15, "Satisfiable"
  | (false, Some _) -> 20, "Theorem"
  | (false, None  ) -> 25, "Unsatisfiable"

let read_thf_file (f:string) (include_fun : string -> unit) =
  let ch = Lexing.from_channel (if (f = "") then stdin else (open_in f)) in
  let old_include_fun = !st_include_fun in
  st_include_fun := include_fun;
(***  List.iter Tptp_config.process_thf (Tptp_parser.tptp_file Tptp_lexer.token ch); ***)
  ignore (Tptp_parser.tptp_file Tptp_lexer.token ch);
  if (!verbosity > 4) then Printf.printf "Finished reading thf file %s\n" f;
  st_include_fun := old_include_fun
      
let rec find_read_thf_file_r odir dir f =
  let ff = (dir ^ "/" ^ f) in
  if (Sys.file_exists ff) then
    read_thf_file ff (find_read_thf_file odir)
  else if (String.length dir > 1) then
    find_read_thf_file_r odir (Filename.dirname dir) f
  else
    raise (FileNotFound f)
and find_read_thf_file dir f =
  let ff = (dir ^ "/" ^ f) in
  if (Sys.file_exists ff) then
    read_thf_file ff (find_read_thf_file dir)
  else
    begin
      try
	let tptpdir = Sys.getenv "TPTP" in
	let tff = (tptpdir ^ "/" ^ f) in
	if (Sys.file_exists tff) then
	  read_thf_file tff (find_read_thf_file dir)
	else
	  find_read_thf_file_r dir dir f
      with
      | Not_found -> find_read_thf_file_r dir dir f
    end;;

st_find_read_thf_fun := find_read_thf_file;;
	
let handle_problem_file () =
  read_thf_file (!problemfile) (find_read_thf_file (Filename.dirname (!problemfile)))

(*** unfinished ***)      
let mizartypechecker_main solv f =
  problemfile := f;
  handle_problem_file ();
  if solv then (*** only try to give an answer to a conjecture if we were asked to ***)
    begin
    end

(*** unfinished ***)      
let mizartypechecker_elaborate f =
  problemfile := f;
  handle_problem_file ();
  ()

let mizartypechecker_elaborate f =
  problemfile := f;
  handle_problem_file ();
  ()
    
let mizartypechecker_extractexamples f =
  problemfile := f;
  handle_problem_file ();
  ()

let number_p x =
  if String.length x = 0 then
    false
  else
    let c = Char.code x.[0] in
    c >= 48 && c < 58
      
let mizar_numbers_hack f i = (*** just add more types I think numbers should satisfy but are not currently derived; at this point they do seem to be derived, but it's still faster to just assume them ***)
  if i = 0 && number_p f then
    begin
      let io = Ar(itp,Prop) in
      let fi = Name(f,itp) in
      let q p = Ap(Name(p,io),fi) in
      let bl = [q "v7_ordinal1";q "v1_int_1";q "v3_ordinal1"] in
      Hashtbl.add mizar_func_types f ("numberhack",0,[],bl)
    end

let load_mizar_knowledge f =
  declare_base_type "$i";
  let c = open_in_bin f in
  mizar_symbol_order_pred := input_value c;
  mizar_name_tp := input_value c;
  mizar_pattern_list := input_value c;
  typingaxiomslist := [];
  try
    while true do
      let mta : mizartypingaxiom = input_value c in
      typingaxiomslist := mta::!typingaxiomslist;
      match mta with
      | MizarFunctionType(ax,fl,i,pl,bl) ->
	 List.iter (fun f -> Hashtbl.add mizar_func_types f (ax,i,pl,bl); mizar_numbers_hack f i) fl
      | MizarTypeHierarchyStrict(ax,fl,i,(p::pl),bl) ->
	 let bl2 = List.filter (fun m -> not (List.mem m (p::pl))) bl in
	 if not (bl2 = []) then
	   begin
	     List.iter (fun f -> Hashtbl.add mizar_typing f (ax,i,p,pl,bl2)) fl
	   end
      | MizarTypeHierarchy(ax,fl,i,(p::pl),bl) ->
	 let bl2 = List.filter (fun m -> not (List.mem m (p::pl))) bl in
	 if not (bl2 = []) then
	   begin
	     List.iter (fun f -> Hashtbl.add mizar_typing f (ax,i,p,pl,bl2)) fl
	   end
      | MizarTypingGeneric(ax,i,pl,ml) ->
	  begin
	   let rec db_occur l i a m =
	     match m with
	     | DB(j,_) when j >= l ->
		let k = j - l in
		if k < i then
		  a.(k) <- true
		else
		  raise Exit
	     | Ap(m1,m2) ->
		db_occur l i a m1;
	       db_occur l i a m2
	     | Lam(_,m1) ->
		db_occur (l+1) i a m1
	     | _ -> ()
	   in
	   let mlsimp =
	     List.filter
	       (fun m -> not (List.mem m pl) &&
		 begin
		   match neg_body m with
		   | Some(nm) ->
		       let (h,s) = head_spine nm in
		       (try ignore (List.find (fun n -> match n with DB(_,_) -> false | _ -> true) s); false with Not_found -> true)
		   | None ->
		       let (h,s) = head_spine m in
		       (try ignore (List.find (fun n -> match n with DB(_,_) -> false | _ -> true) s); false with Not_found -> true)
		 end)
	       ml
	   in
	   if not (mlsimp = []) then
	     begin
	       let pl1 = ref pl in
	       let pr1 = ref [] in
	       while not (!pl1 = []) do
		 match !pl1 with
		 | [] -> () (*** impossible ***)
		 | (prim::plr) ->
		    pl1 := plr;
		   let alldboccur =
		     begin
		       let a = Array.create i false in
		       try
			 db_occur 0 i a prim;
			 for j = 0 to i-1 do
			   if not a.(j) then raise Exit
			 done;
			 true
		       with Exit -> false
		     end
		   in
		   if alldboccur then
		     begin
		       match neg_body prim with
		       | Some(nprim) ->
			  begin
			    let (h,s) = head_spine nprim in
			    match h with
			    | Name(f,_) -> Hashtbl.add mizar_typing_gen_forward f (ax,i,prim,(!pr1 @ !pl1),mlsimp)
			    | _ -> ()
			  end
		       | None ->
			  begin
			    let (h,s) = head_spine prim in
			    match h with
			    | Name(f,_) -> Hashtbl.add mizar_typing_gen_forward f (ax,i,prim,(!pr1 @ !pl1),mlsimp)
			    | _ -> ()
			  end
		     end;
		   pr1 := prim::!pr1
	       done
	     end;
	   List.iter
	     (fun concl ->
	       if not (List.mem concl pl) then
		 let alldboccur =
		   begin
		     let a = Array.create i false in
		     try
		       db_occur 0 i a concl;
		       for j = 0 to i-1 do
			 if not a.(j) then raise Exit
		       done;
		       true
		     with Exit -> false
		   end
		 in
		 if alldboccur then
		   begin
		     match neg_body concl with
		     | Some(nconcl) ->
			begin
			  let (mh,ms) = head_spine nconcl in
			  match ms with
			  | [] -> ()
			  | (a::r) ->
			     let (h,s) = head_spine a in
			     match h with
			     | Name(f,_) -> Hashtbl.add mizar_typing_gen_backward f (ax,i,a,pl,concl)
			     | _ -> ()
			end
		     | None ->
			let (mh,ms) = head_spine concl in
			match ms with
			| [] -> ()
			| (a::r) ->
			   let (h,s) = head_spine a in
			   match h with
			   | Name(f,_) -> Hashtbl.add mizar_typing_gen_backward f (ax,i,a,pl,concl)
			   | _ -> ()
		   end)
	     ml
	 end
      | MizarRedefinition(ax,i,pl,f,s,d) ->
	 Hashtbl.add mizar_redef f (ax,i,pl,s,d)
      | _ -> ()
    done
  with End_of_file ->
    close_in c;
    List.iter
      (fun (nname,i,pl,rtp,al,pd,d) ->
	Hashtbl.add mizar_pattern_table nname (i,pl,rtp,al,pd,d))
      !mizar_pattern_list
      
type gentp = Stp of stp | Mtp of trm list

type mizrecinfo = ((string list * bool) list) * ((string * trm) list * (trm * gentp * (string * trm list) list)) list
	
let rec list_of_strings_of_formal_reps_of_rich_grammar_tree_type (tr:(identifier,mizrecinfo option) rich_grammar_tree_type) =
  match tr with
  | RNT(i,r,cl) ->
      begin
	match !r with
	| None ->
	    []
	| Some(_,rl) ->
	    let fst = ref true in
	    List.map
	      (fun (_,(n,_,_)) ->
		try
		  let (f,_) = trm_to_foform_stp (onlybetanorm (logicnorm n)) true in
		  trm_str_nice n ^ ";FOFORM:" ^ printf_fof f
		with NotFO ->
		  trm_str_nice n ^ ";NotFO")
	      rl;
      end
  | RT(i,r) ->
      begin
	match !r with
	| None ->
	    []
	| Some(_,rl) ->
	    let fst = ref true in
	    List.map
	      (fun (_,(n,_,_)) -> trm_str_nice n)
	      rl;
      end

let rec string_of_rich_grammar_tree_type_rec b (tr:(identifier,mizrecinfo option) rich_grammar_tree_type) =
  match tr with
  | RNT(i,r,cl) ->
     begin
       Buffer.add_string b (Printf.sprintf "(RNT \"%s\"" (string_of_identifier i));
       begin
	 match !r with
	 | None ->
	    Buffer.add_string b " []"
	 | Some(ul,rl) ->
	     let fst = ref true in
	     Buffer.add_string b " [[";
	     List.iter
	       (fun (_,_) ->
		 Buffer.add_string b "\"something";
		 Buffer.add_string b "\"")
	       ul;
	     Buffer.add_string b "];[";
	     List.iter
	       (fun (_,(n,_,_)) ->
		 if !fst then fst := false else Buffer.add_string b ";";
		 Buffer.add_string b "\"";		 
		 Buffer.add_string b (trm_str_nice n);
		 Buffer.add_string b "\"")
	       rl;
	     Buffer.add_string b "]]"
       end;
       List.iter
	 (fun c ->
	   Buffer.add_string b " ";
	   string_of_rich_grammar_tree_type_rec b c)
	 cl;
       Buffer.add_string b ")"
     end
  | RT(i,r) ->
     begin
       Buffer.add_string b (Printf.sprintf "(RT \"%s\"" (string_of_identifier i));
       begin
	 match !r with
	 | None ->
	    Buffer.add_string b " []"
	 | Some(_,rl) ->
	     let fst = ref true in
	     Buffer.add_string b "[";
	     List.iter
	       (fun (_,(n,_,_)) ->
		 if !fst then fst := false else Buffer.add_string b ";";
		 Buffer.add_string b "\"";
		 Buffer.add_string b (trm_str_nice n);
		 Buffer.add_string b "\"")
	       rl;
	     Buffer.add_string b "]"
       end;
       Buffer.add_string b ")"
     end

let string_of_rich_grammar_tree_type tr =
  let b = Buffer.create 100 in
  string_of_rich_grammar_tree_type_rec b tr;
  Buffer.contents b

let string_of_rich_grammar_tree_type_l trl =
  let b = Buffer.create 100 in
  List.iter (fun tr -> string_of_rich_grammar_tree_type_rec b tr; Buffer.add_char b ';') trl;
  Buffer.contents b

exception BadParseException
exception DontKnowException of string
    
type mizar_elab_res =
    BadParse
  | MizarTimeout
  | IncompleteArgs
  | CheckingFailure
  | ElaborationSuccess
  | DontKnow

let mer_string mer =
  match mer with
  | BadParse -> "BadParse"
  | MizarTimeout -> "MizarTimeout"
  | IncompleteArgs -> "IncompleteArgs"
  | CheckingFailure -> "CheckingFailure"
  | ElaborationSuccess -> "ElaborationSuccess"
  | DontKnow -> "DontKnow"

let rec genap f al =
  match al with
  | [] -> f
  | (a::ar) -> genap (Ap(f,a)) ar
     
let rec genpreap f al =
  match al with
  | [] -> f
  | (a::ar) -> genpreap (PAp(f,a)) ar
     
let stp_to_gentp st m a =
  match a with
  | Base("$i") ->
     let (_,_,sh,_,_,_) = st in
     Mtp(Hashtbl.find_all sh m)
  | _ -> Stp(a)

let rec extract_children trl tl gtl acl oth =
  match trl with
  | [] -> (List.rev tl,List.rev gtl,List.rev acl,List.rev oth)
  | tr::trr ->
      match tr with
      | RT(i,_) when List.mem i [globalids.lparen;globalids.rparen;globalids.lparenhash;globalids.rparenhash;globalids.comma] -> extract_children trr tl gtl acl (tr::oth)
      | RT(_,_) -> extract_children trr (tr::tl) gtl acl oth
      | RNT(i,_,_) when i = globalids.ddcomma -> extract_children trr tl gtl acl (tr::oth)
      | RNT(i,_,_) when i = globalids.ddof || i = globalids.ddover || i = globalids.ddnot -> extract_children trr tl gtl acl (tr::oth)
      | RNT(i,_,[RT(_,_)]) when not (i = globalids.numi) -> extract_children trr tl (tr::gtl) acl oth
      | _ -> extract_children trr tl gtl (tr::acl) oth

let allowed_aux_p k = List.mem k [globalids.ddthe;globalids.ddof;globalids.ddover;globalids.ddnot;globalids.ddnon]

let rec extract_genterm_head_name_3 i j gtl quiet aux =
  match gtl with
  | [] -> (string_of_identifier j,aux)
  | (RNT(k,_,[RT(l,_)])::gtr) ->
      if allowed_aux_p k then
	extract_genterm_head_name_3 i j gtr quiet (k::aux)
      else
	extract_genterm_head_name_3 i j gtr quiet aux
      (***
      if k = i && l = j then
	extract_genterm_head_name_3 i j gtr quiet aux
      else if allowed_aux_p k then
	extract_genterm_head_name_3 i j gtr quiet (k::aux)
      else
	begin
	  if not quiet then Printf.printf "#+#++ ERROR: (1) More than one head given by generalized terminals: %s and %s\n" (string_of_identifier i) (string_of_identifier j);
	  raise BadParseException
	end
	 ***)
  | (_::_) -> raise (Failure("list of generalized terminals contained something other than a generalized terminal"))

let rec extract_genterm_head_name_2 i gtl quiet aux =
  match gtl with
  | [] -> (string_of_identifier i,aux)
  | (RNT(j,r,cl)::gtr) ->
      if allowed_aux_p j then
	extract_genterm_head_name_2 i gtr quiet (j::aux)
      else
	extract_genterm_head_name_2 i gtr quiet aux
	  (***
      if j = i then
	extract_genterm_head_name_2 i gtr quiet aux
      else if allowed_aux_p j then
	extract_genterm_head_name_2 i gtr quiet (j::aux)
      else
	begin
	  if not quiet then Printf.printf "#+#++ ERROR: (2) More than one head given by generalized terminals: %s and %s\n" (string_of_identifier i) (string_of_identifier j);
	  raise BadParseException
	end
***)
  | (_::_) -> raise (Failure("list of generalized terminals contained something other than a generalized terminal"))

let lasttr = ref (RT(0,ref None))
    
let rec extract_genterm_head_name_r gtl quiet aux =
  match gtl with
  | [] when List.mem globalids.ddcomma aux ->
      Printf.printf "#+#++ TODO: comma\n";
      raise BadParseException
  | [] ->
      if not quiet then Printf.printf "#+#++ ERROR: (1) Could not find any generalized terminals at all to get a head: %s\n" (string_of_rich_grammar_tree_type !lasttr);
      raise BadParseException
  | (RNT(i,_,[RT(j,_)])::gtr) ->
      if string_of_identifier i = "Var" then
	begin
	  let (x,aux) = extract_genterm_head_name_3 i j gtr quiet aux in
	  match aux with
	  | [] -> (x,true)
	  | _ ->
	      if not quiet then (Printf.printf "#+#++ Var unhandled aux case"; List.iter (fun j -> Printf.printf " %s" (string_of_identifier j)) aux; Printf.printf "\n");
	      raise BadParseException
	end
      else if allowed_aux_p i then
	extract_genterm_head_name_r gtr quiet (i::aux)
      else
	begin
	  let (h,aux) = extract_genterm_head_name_2 i gtr quiet aux in
	  if String.length h > 2 && h.[0] = '$' && h.[1] = '#' then
	    begin
	      match aux with
	      | [] -> (String.sub h 2 (String.length h - 2),false)
	      | [z] when z = globalids.ddthe ->
		  let x = String.sub h 2 (String.length h - 2) in (*** ignore 'the' in this case ***)
		  (x,false)
	      | _ ->
		  if not quiet then (Printf.printf "#+#++ Nonvar unhandled aux case"; List.iter (fun j -> Printf.printf " %s" (string_of_identifier j)) aux; Printf.printf "\n");
		  raise BadParseException
	    end
	  else
	    begin
	      if not quiet then Printf.printf "#+#++ Surpising head name %s not starting with $#\n" h;
	      (h,false)
	    end
	end
  | (_::_) -> raise (Failure("list of generalized terminals contained something other than a generalized terminal"))

let extract_genterm_head_name gtl quiet aux =
  try
    extract_genterm_head_name_r gtl quiet aux
  with BadParseException ->
    if not quiet then
      begin
	Printf.printf "#+#++ extract_genterm_head_name called with %d:\n" (List.length gtl);
	List.iter (fun gt -> Printf.printf "#+#++ %s\n" (string_of_rich_grammar_tree_type gt)) gtl;
      end;
    raise BadParseException
	
let rec extract_varlist_l cl xl =
  if debug then Printf.printf "#+#++ extract_varlist_l %s\n" (string_of_rich_grammar_tree_type_l cl);
  match cl with
  | RNT(i,_,[RNT(j,_,[RT(k,_)])])::cr when i = globalids.seti && j = globalids.var ->
      extract_varlist_l cr (string_of_identifier k::xl)
  | RNT(i,_,cl1)::cr when i = globalids.varlist ->
      extract_varlist_l cr (extract_varlist_l cl1 xl)
  | RNT(i,_,_)::cr when i = globalids.ddcomma ->
      extract_varlist_l cr xl
  | RT(i,_)::cr when i = globalids.comma ->
      extract_varlist_l cr xl
  | _ -> xl

let extract_varlist tr xl =
  if debug then Printf.printf "#+#++ extract_varlist %s\n" (string_of_rich_grammar_tree_type tr);
  match tr with
  | RNT(i,_,cl) when i = globalids.varlist -> extract_varlist_l cl xl
  | _ -> xl

let typish_id_p i =
  if i = globalids.typei || i = globalids.attypei || i = globalids.atlistofadjectives || i = globalids.attribute || i = globalids.adjective then
    true
  else
    let h = string_of_identifier i in
    String.length h > 3 && h.[0] = '$' && h.[1] = '#' && h.[2] = '~'

type rgt = (identifier,mizrecinfo option) rich_grammar_tree_type
type frctx = (string list * rgt option) list
      
let extract_fraenkel_parts_6 (cl:rgt list) xlvtpl body cond : frctx * rgt * rgt option =
  if debug then Printf.printf "efp 6 %s\n" (string_of_rich_grammar_tree_type_l cl);
  match cl with
  | [RNT(i,_,_)] when i = globalids.ddfraenkelrbrace ->
      (xlvtpl,body,cond)
  | [RT(i,_)] when string_of_identifier i = "}" -> (*** delete after fixed ***)
      Printf.printf "#+#++ ** WARNING ** }\n";
      (xlvtpl,body,cond)
  | _ -> raise BadParseException
  
let extract_fraenkel_parts_5 (cl:rgt list) xlvtpl body : frctx * rgt * rgt option =
  if debug then Printf.printf "efp 5 %s\n" (string_of_rich_grammar_tree_type_l cl);
  match cl with
  | ((RNT(i,_,_) as tr)::cr) when i = globalids.booli ->
      extract_fraenkel_parts_6 cr xlvtpl body (Some(tr))
  | _ -> raise BadParseException
  
let extract_fraenkel_parts_4 (cl:rgt list) xlvtpl body : frctx * rgt * rgt option =
  if debug then Printf.printf "efp 4 %s\n" (string_of_rich_grammar_tree_type_l cl);
  match cl with
  | (RNT(i,_,_)::cr) when i = globalids.ddfraenkelcolon ->
      extract_fraenkel_parts_5 cr xlvtpl body
  | (RT(i,_)::cr) when string_of_identifier i = ":" -> (** delete after fixed **)
      Printf.printf "#+#++ ** WARNING ** :\n";
      extract_fraenkel_parts_5 cr xlvtpl body
  | _ -> extract_fraenkel_parts_6 cl xlvtpl body None
  
let rec extract_fraenkel_parts_3 (clv:rgt list) xlvtpr (cl:rgt list) body : frctx * rgt * rgt option =
  if debug then Printf.printf "efp 3 %s\n" (string_of_rich_grammar_tree_type_l clv);
  match clv with
  | (RNT(i,_,cr2)::cr) when i = globalids.fraenkelvars ->
      extract_fraenkel_parts_3 (cr2 @ cr) xlvtpr cl body (*** questionable @ ***)
  | ((RNT(i,_,_) as tr)::crv) when i = globalids.varlist ->
      let xl = extract_varlist tr [] in
      extract_fraenkel_parts_3b crv xl xlvtpr cl body
  | (RNT(i,_,_)::crv) when i = globalids.ddcomma ->
      extract_fraenkel_parts_3 crv xlvtpr cl body
  | (RT(i,_)::crv) when i = globalids.comma -> (** can delete after generalized, but not vital **)
      extract_fraenkel_parts_3 crv xlvtpr cl body
  | [] ->
      extract_fraenkel_parts_4 cl (List.rev xlvtpr) body
  | _ -> raise BadParseException
and extract_fraenkel_parts_3b (clv:rgt list) xl xlvtpr (cl:rgt list) body : frctx * rgt * rgt option =
  if debug then Printf.printf "efp 3b %s\n" (string_of_rich_grammar_tree_type_l clv);
  match clv with
  | (RNT(i,_,_)::crv) when i = globalids.ddis -> extract_fraenkel_parts_3b crv xl xlvtpr cl body
  | ((RNT(i,_,_) as tr)::crv) when typish_id_p i ->
      extract_fraenkel_parts_3 crv ((xl,Some(tr))::xlvtpr) cl body
  | [] ->
      extract_fraenkel_parts_4 cl (List.rev ((xl,None)::xlvtpr)) body
  | _ -> raise BadParseException

let extract_fraenkel_parts_2 (cl:rgt list) body : frctx * rgt * rgt option =
  if debug then Printf.printf "efp 2 %s\n" (string_of_rich_grammar_tree_type_l cl);
  match cl with
  | (RNT(i,_,cr2)::cr) when i = globalids.fraenkelvars ->
      extract_fraenkel_parts_3 cr2 [] cr body
  | _ -> raise BadParseException

let extract_fraenkel_parts_1b (cl:rgt list) body : frctx * rgt * rgt option =
  if debug then Printf.printf "efp 1b %s\n" (string_of_rich_grammar_tree_type_l cl);
  match cl with
  | (RNT(i,_,_)::cr) when i = globalids.ddfraenkelwhere ->
      extract_fraenkel_parts_2 cr body
  | (RT(i,_)::cr) when string_of_identifier i = "where" -> (*** delete after fixed ***)
      Printf.printf "#+#++ ** WARNING ** where\n";
      extract_fraenkel_parts_2 cr body
  | _ -> raise BadParseException
	
let extract_fraenkel_parts_1 (cl:rgt list) : frctx * rgt * rgt option =
  if debug then Printf.printf "efp 1 %s\n" (string_of_rich_grammar_tree_type_l cl);
  match cl with
  | (RNT(i,_,_) as tr)::cr when i = globalids.seti ->
      extract_fraenkel_parts_1b cr tr
  | _ -> raise BadParseException

let extract_fraenkel_parts (cl:rgt list) : frctx * rgt * rgt option =
  match cl with
  | (RNT(i,_,_) as tr)::cr when i = globalids.ddfraenkellbrace ->
      extract_fraenkel_parts_1 cr
  | RT(i,_)::cr when string_of_identifier i = "{" -> (*** delete after fixed ***)
      Printf.printf "#+#++ ** WARNING ** {\n";
      extract_fraenkel_parts_1 cr
  | _ -> raise BadParseException

let extract_univ_parts_4 cl xl vtp suchth =
  if debug then Printf.printf "eup 4 %s\n" (string_of_rich_grammar_tree_type_l cl);
  match cl with
  | [tr] -> (xl,vtp,suchth,tr)
  | _ -> raise BadParseException
  
let extract_univ_parts_3 cl xl vtp =
  if debug then Printf.printf "eup 3 %s\n" (string_of_rich_grammar_tree_type_l cl);
  match cl with
  | (tr::RNT(i,_,_)::cr) when i = globalids.ddholds ->
      extract_univ_parts_4 cr xl vtp (Some(tr))
  | _ -> raise BadParseException

let extract_univ_parts_2 cl xl =
  if debug then Printf.printf "eup 2 %s\n" (string_of_rich_grammar_tree_type_l cl);
  match cl with
  | (tr::RNT(i,_,_)::cr) when i = globalids.ddst ->
      extract_univ_parts_3 cr xl (Some(tr))
  | (tr::RNT(i,_,_)::cr) when i = globalids.ddholds ->
      extract_univ_parts_4 cr xl (Some(tr)) None
  | [tr;tr2] -> (xl,Some(tr),None,tr2) (*** if there is no explicit st or holds, then treat tr2 as the 'holds' ***)
  | _ -> raise BadParseException
	
let rec extract_univ_parts_1 cl xl =
  if debug then (Printf.printf "eup 1 %s\n" (string_of_rich_grammar_tree_type_l cl); List.iter (fun x -> Printf.printf " %s\n" x) xl);
  match cl with
  | (RNT(i,_,_) as tr)::cr when i = globalids.varlist ->
      extract_univ_parts_1 cr (extract_varlist tr xl)
  | RNT(i,_,_)::cr when i = globalids.ddbeing ->
      extract_univ_parts_2 cr (List.rev xl)
  | RNT(i,_,_)::cr when i = globalids.ddst ->
      extract_univ_parts_3 cr (List.rev xl) None
  | RNT(i,_,_)::cr when i = globalids.ddholds ->
      extract_univ_parts_4 cr (List.rev xl) None None
  | [tr] -> (xl,None,None,tr) (*** if there is no explicit st or holds, then treat tr2 as the 'holds' ***)
  | _ -> raise BadParseException

let extract_univ_parts cl =
  match cl with
  | (RNT(i,_,_)::cr) when i = globalids.ddfor -> extract_univ_parts_1 cr []
  | _ -> raise BadParseException

let extract_ex_parts_3 cl xl vtp =
  if debug then Printf.printf "eep 3 %s\n" (string_of_rich_grammar_tree_type_l cl);
  match cl with
  | [tr] -> (xl,vtp,tr)
  | _ -> raise BadParseException

let extract_ex_parts_2 cl xl =
  if debug then Printf.printf "eep 2 %s\n" (string_of_rich_grammar_tree_type_l cl);
  match cl with
  | (tr::RNT(i,_,_)::cr) when i = globalids.ddst ->
      extract_ex_parts_3 cr xl (Some(tr))
  | [tr;tr2] -> (xl,Some(tr),tr2) (*** if there is no explicit st, then treat tr2 as the 'holds' ***)
  | _ -> raise BadParseException
	
let rec extract_ex_parts_1 cl xl =
  if debug then Printf.printf "eep 1 %s\n" (string_of_rich_grammar_tree_type_l cl);
  match cl with
  | (RNT(i,_,_) as tr)::cr when i = globalids.varlist ->
      extract_ex_parts_1 cr (extract_varlist tr xl)
  | RNT(i,_,_)::cr when i = globalids.ddbeing ->
      extract_ex_parts_2 cr (List.rev xl)
  | RNT(i,_,_)::cr when i = globalids.ddst ->
      extract_ex_parts_3 cr (List.rev xl) None
  | [tr] -> (xl,None,tr) (*** if there is no explicit st, then treat tr2 as the 'st' ***)
  | _ -> raise BadParseException

let extract_ex_parts cl =
  match cl with
  | (RNT(i,_,_)::cr) when i = globalids.ddex -> extract_ex_parts_1 cr []
  | _ -> raise BadParseException

let lookup_precomp (g:(string * trm) list) (r:mizrecinfo option ref) : (trm * gentp * (string * trm list) list) option =
  match !r with
  | None -> None
  | Some(_,a) ->
      try
	let (n,ntp,frl) = List.assoc g a in
	Some(n,ntp,frl)
      with Not_found -> None

let add_precomp (g:(string * trm) list) (r:mizrecinfo option ref) (v: trm * gentp * (string * trm list) list) : unit =
  match !r with
  | None -> r := Some([],[(g,v)])
  | Some(b,a) -> r := Some(b,((g,v)::a))

let lookup_unknown_precomp (g:string list) (r:mizrecinfo option ref) : bool option =
  match !r with
  | None -> None
  | Some(a,_) ->
      try
	let b = List.assoc g a in
	Some(b)
      with Not_found -> None

let add_unknown_precomp (g:string list) (r:mizrecinfo option ref) (b:bool) : bool =
  match !r with
  | None -> r := Some([(g,b)],[]); b
  | Some(a1,a2) -> r := Some(((g,b)::a1),a2); b

	(*** quick check fo unknown names in the parse tree ***)
let rec unknown_name_occurs_p g tr =
  lasttr := tr;
  if debug then
    begin
      Printf.printf "unknown_name_occurs_p Input: %s\n" (string_of_rich_grammar_tree_type tr);
      Printf.printf "g:";
      List.iter (fun y -> Printf.printf " %s" y) g;
      Printf.printf "\n";
    end;
  let b = unknown_name_occurs_p_r g tr in
  if debug then
    begin
      Printf.printf "unknown_name_occurs_p Output %b Input: %s\n" b (string_of_rich_grammar_tree_type tr);
      Printf.printf "g:";
      List.iter (fun y -> Printf.printf " %s" y) g;
      Printf.printf "\n";
    end;
  b
and unknown_name_occurs_p_r g tr =
  match tr with
  | RNT(i,r,cl) ->
      begin
	match lookup_unknown_precomp g r with
	| Some(b) -> b
	| _ ->
	    if i = globalids.fraenkel then
	      add_unknown_precomp g r (unknown_name_occurs_fraenkel g cl)
	    else
	      let (tl,gtl,acl,oth) = extract_children cl [] [] [] [] in
	      match (gtl,acl) with
	      | (RNT(j,_,_)::_,_) when j = globalids.ddfor -> add_unknown_precomp g r (unknown_name_occurs_univ g cl)
	      | (RNT(j,_,_)::_,_) when j = globalids.ddex -> add_unknown_precomp g r (unknown_name_occurs_ex g cl)
	      | (RNT(j,_,_)::_,_) when List.mem j [globalids.ddverum;globalids.ddamp;globalids.ddor;globalids.ddimplies;globalids.ddiff;globalids.ddif;globalids.ddotherwise;globalids.ddnot;globalids.ddnon;globalids.ddst;globalids.ddholds;globalids.ddbe;globalids.ddbeing;globalids.ddis;globalids.ddof;globalids.ddthe;globalids.ddover;globalids.seti;globalids.booli;globalids.typei;globalids.attypei] ->
		  add_unknown_precomp g r (unknown_name_occurs_l_p g acl)
	      | _ ->
		  try
	      (*** abusing BadParseException to mean 'check the children for unknowns' here ***)
		    let (h,v) = extract_genterm_head_name gtl true [] in
		    if v then
		      if List.mem h g then
			raise BadParseException
		      else
			begin
			  Printf.printf "#+#++ Unknown name %s\n" h;
			  add_unknown_precomp g r true
			end
		    else
		      if Hashtbl.mem !mizar_name_tp h then
			raise BadParseException
		      else
			begin
			  Printf.printf "#+#++ Unknown name %s\n" h;
			  add_unknown_precomp g r true
			end
    		  with BadParseException ->
		    add_unknown_precomp g r (unknown_name_occurs_l_p g acl)
      end
  | RT(i,r) -> false
and unknown_name_occurs_l_p g cl =
  if debug then
    begin
      Printf.printf "unknown_name_occurs_l_p Input: %s\n" (string_of_rich_grammar_tree_type_l cl);
      Printf.printf "g:";
      List.iter (fun y -> Printf.printf " %s" y) g;
      Printf.printf "\n";
    end;
  match cl with
  | [] -> false
  | (tr::cr) -> unknown_name_occurs_p g tr || unknown_name_occurs_l_p g cr
and unknown_name_occurs_univ g cl =
  let (xl,vtp,suchth,body) = extract_univ_parts cl in
  try
    begin
      match vtp with
      | Some(vtp) -> if unknown_name_occurs_p g vtp then raise Exit
      | None -> ()
    end;
    let g2 = xl @ g in
    begin
      match suchth with
      | Some(suchth) -> if unknown_name_occurs_p g2 suchth then raise Exit
      | _ -> ()
    end;
    unknown_name_occurs_p g2 body
  with Exit -> true
and unknown_name_occurs_ex g cl =
  let (xl,vtp,body) = extract_ex_parts cl in
  try
    begin
      match vtp with
      | Some(vtp) -> if unknown_name_occurs_p g vtp then raise Exit
      | None -> ()
    end;
    let g2 = xl @ g in
    unknown_name_occurs_p g2 body
  with Exit -> true
and unknown_name_occurs_fraenkel g cl =
  if debug then Printf.printf "unknown_name_occurs_fraenkel %s\n" (string_of_rich_grammar_tree_type_l cl);
  let (xlvtpl,body,cond) = extract_fraenkel_parts cl in
  try
    List.iter
      (fun (_,vtp) ->
	begin
	  match vtp with
	  | Some(vtp) -> if unknown_name_occurs_p g vtp then raise Exit
	  | None -> ()
	end)
      xlvtpl;
    let g2 = List.fold_left (fun gi (xl,_) -> xl @ gi) g xlvtpl in
    if unknown_name_occurs_p g2 body then raise Exit;
    match cond with
    | Some(cond) ->
	if unknown_name_occurs_p g2 cond then raise Exit;
	false
    | None -> false
  with Exit -> true

let rec fix_parse_tree_l cl =
  match cl with
  | ((RNT(i,_,_) as tr1)::(RNT(j,_,_) as tr2)::cr) when i = globalids.ddthe && j = globalids.typei ->
      RNT(globalids.seti,ref None,[tr1;tr2])::cr
  | tr1::cr -> tr1::fix_parse_tree_l cr
  | _ -> raise Not_found
	
let fix_parse_tree tr =
  match tr with
(*  | RNT(i,r,[(RNT(j0,_,_) as tr0);(RNT(j1,_,_) as tr1);(RT(j2,_) as tr2);(RNT(j3,_,_) as tr3);(RNT(j4,_,_) as tr4);(RNT(j5,_,_) as tr5);(RT(j6,_) as tr6)])
    when j2 = globalids.lparen && j6 = globalids.rparen ->
      Printf.printf "#+#++ WARNING: Parse tree has internal parens. Is it OK? %s\n" (string_of_rich_grammar_tree_type tr);
      raise Not_found *)
(*      RNT(i,r,[tr0;tr1;RNT(i,ref None,[tr3;tr4;tr5])]) *)
  | RNT(i,r,(tr1::cr)) ->
      let cr2 = fix_parse_tree_l cr in
      RNT(i,r,tr1::cr2)
  | _ -> raise Not_found
	
let rec mizar_elab_and_test_rec st g (tr:(identifier,mizrecinfo option) rich_grammar_tree_type) =
  lasttr := tr;
  if debug then
    begin
      Printf.printf "rec Input: %s\n" (string_of_rich_grammar_tree_type tr);
      Printf.printf "g:\n";
      List.iter (fun (y,n) -> Printf.printf "%s ~> %s\n" y (trm_str n)) g
    end;
  try
    let trfix = fix_parse_tree tr in
    mizar_elab_and_test_rec st g trfix
  with Not_found ->
    match tr with
    | RNT(i,r,cl) ->
	begin
	  if debug then Printf.printf "#+#++ 0 mizar_elab_and_test_rec RNT %d %s\n" i (string_of_identifier i); flush stdout;
	  match lookup_precomp g r with
	  | Some(m,gtp,frl) -> (m,frl)
	  | None ->
	      if debug then Printf.printf "#+#++ mizar_elab_and_test_rec RNT %d %s\n" i (string_of_identifier i); flush stdout;
	      if i = globalids.fraenkel then
		let (n,frl) = mizar_elab_and_test_fraenkel r st g cl in
		add_precomp g r (n,Mtp([Ap(Name("m1_hidden",Ar(itp,Prop)),n)]),frl);
		(n,frl)
	      else if i = globalids.booli then
		match cl with
		| [RT(i,_);(RNT(_,r1,_) as tr1);RT(j,_)] when i = globalids.lparen && j = globalids.rparen ->
		    begin
		      let n = mizar_elab_and_test_rec st g tr1 in
		      r := !r1;
		      n
		    end
		| _ ->
		    let (tl,gtl,acl,oth) = extract_children cl [] [] [] [] in
		    if not (tl = []) then
		      begin
			Printf.printf "#+#++ Unexpected terminal children %d %s\n" (List.length tl) (string_of_rich_grammar_tree_type tr);
			raise BadParseException
		      end;
		    if debug then
		      begin
			Printf.printf "#+#++ Bool node with gen terminals %s\n" (string_of_rich_grammar_tree_type_l gtl);
		      end;
		    if not (List.length gtl = 1) then
		      begin
			match (gtl,acl) with
			| ([],[RNT(j,rj,_) as trj]) when i = j -> ()
			| (RNT(j,_,_)::_,_) when j = globalids.ddfor -> ()
			| (RNT(j,_,_)::_,_) when j = globalids.ddex -> ()
			| (RNT(j,_,_)::_,_) when j = globalids.ddst -> ()
			| (RNT(j,_,_)::_,_) when j = globalids.ddholds -> ()
			| _ ->
			    Printf.printf "#+#++ ERROR: not = 1 but %d: %s ; the node: %s\n" (List.length gtl) (string_of_rich_grammar_tree_type_l gtl) (string_of_rich_grammar_tree_type tr);
			    if debug then
			      List.iteri
				(fun i z ->
				  Printf.printf "child %d. %s\n" i
				    (match z with
				    | RT(j,_) -> string_of_identifier j
				    | RNT(j,_,[RT(k,_)]) -> string_of_identifier j ^ "[" ^ string_of_identifier k ^ "]"
				    | _ -> "*"))
				cl;
		      end;
		    let negated = try ignore (List.find (fun trj -> match trj with RNT(j,_,_) when j = globalids.ddnot -> true | _ -> false) oth); true with Not_found -> false in (*** treat arbitrary number of nots at a single parse tree level as a single not; obviously questionable, but if the intention was that there are really multiple negations then they should each be parsed at their own level; multiple nots at a single level shouldn't really be happening ***)
		    if debug then Printf.printf "#+#++ negated %b oth %s\n" negated (string_of_rich_grammar_tree_type_l oth);
		    match (gtl,acl) with
		    | ([],[RNT(j,rj,_) as trj]) when i = j && not negated -> (*** intended as casting ***)
			begin
			  let n = mizar_elab_and_test_rec st g trj in
			  r := !rj;
			  n
			end
		    | (RNT(j,_,_)::_,_) when j = globalids.ddfor -> mizar_elab_and_test_all r st g cl negated
		    | (RNT(j,_,_)::_,_) when j = globalids.ddex -> mizar_elab_and_test_ex r st g cl negated
		    | (RNT(j,_,_)::_,_) when j = globalids.ddof -> raise BadParseException
		    | (RNT(j,_,_)::_,_) when j = globalids.ddis ->
			begin
			  match acl with
			  | [tr1;tr2] ->
			      if debug then Printf.printf "#+#++ is %s  ;  %s\n" (string_of_rich_grammar_tree_type tr1) (string_of_rich_grammar_tree_type tr2);
			      let (n1,frl1) = mizar_elab_and_test_rec st g tr1 in
			      if debug then Printf.printf "#+#++ is n1 %s\n" (trm_str n1);
			      let (n2,frl2) = mizar_elab_and_test_sort st g tr2 in
			      if debug then Printf.printf "#+#++ is n2 %s\n" (trm_str n2);
			      let frl = union_frees frl1 frl2 in
			      if tpof n1 = itp then
				begin
				  if tpof n2 = Prop then
				    begin
				      let n = subst n2 0 n1 in
				      let n = if negated then neg n else n in
				      add_precomp g r (n,Stp(Prop),frl);
				      (n,frl)
				    end
				  else
				    raise (ExpectedTypeError(PTrm(n2),tpof n2,Prop))
				end
			      else
				raise (ExpectedTypeError(PTrm(n1),tpof n1,itp))
			  | _ -> raise BadParseException
			end
		    | (RNT(j,_,_)::_,_) when j = globalids.ddbe -> raise BadParseException
		    | (RNT(j,_,_)::_,_) when j = globalids.ddbeing -> raise BadParseException
		    | (RNT(j,_,_)::_,_) when j = globalids.ddst -> (*** this should be part of a $$for/universal or $$ex/existential, parsed in a way I didn't expect but may be acceptable ***)
			begin
			  match cl with
			  | [RNT(j,_,_);tr1;RNT(k,_,_);tr2] when j = globalids.ddst && k = globalids.ddholds ->
			      mizar_elab_and_test_logbinop st g [tr1;tr2] r imp negated
			  | [RNT(k,_,_);(RNT(j,rj,_) as tr2)] when k = globalids.ddst ->
			      begin
				let (n,frl) = mizar_elab_and_test_rec st g tr2 in
				let n = if negated then neg n else n in
				add_precomp g r (n,Stp(Prop),frl);
				(n,frl)
			      end
			  | _ ->
			      raise BadParseException
			end
		    | (RNT(j,_,_)::_,_) when j = globalids.ddholds -> (*** this should be part of a $$for/universal, parsed in a way I didn't expect but may be acceptable ***)
			begin
			  match cl with
			  | [RNT(j,_,_);tr1;RNT(k,_,_);tr2] when j = globalids.ddst && k = globalids.ddholds ->
			      mizar_elab_and_test_logbinop st g [tr1;tr2] r imp negated
			  | [RNT(k,_,_);(RNT(j,rj,_) as tr2)] when k = globalids.ddholds ->
			      begin
				let (n,frl) = mizar_elab_and_test_rec st g tr2 in
				let n = if negated then neg n else n in
				add_precomp g r (n,Stp(Prop),frl);
				(n,frl)
			      end
			  | _ ->
			      raise BadParseException
			end
		    | (RNT(j,_,_)::_,_) when j = globalids.ddnon -> raise BadParseException
		    | ([],_) when negated ->
			begin
			  match acl with
			  | [tr1] ->
			      let (n1,frl1) = mizar_elab_and_test_rec st g tr1 in
			      if tpof n1 = Prop then
				begin
				  let n = neg n1 in
				  add_precomp g r (n,Stp(Prop),frl1);
				  (n,frl1)
				end
			      else
				raise (ExpectedTypeError(PTrm(n1),tpof n1,Prop))
			  | _ -> raise BadParseException
			end
		    | (RNT(j,_,_)::_,_) when j = globalids.ddverum ->
			if negated then
			  (False,[])
			else
			  (Ap(Neg,False),[])
		    | (RNT(j,_,_)::_,_) when j = globalids.ddamp ->
			mizar_elab_and_test_logbinop st g acl r conj negated
		    | (RNT(j,_,_)::_,_) when j = globalids.ddor ->
			mizar_elab_and_test_logbinop st g acl r disj negated
		    | (RNT(j,_,_)::_,_) when j = globalids.ddimplies ->
			mizar_elab_and_test_logbinop st g acl r imp negated
		    | (RNT(j,_,_)::_,_) when j = globalids.ddiff ->
			mizar_elab_and_test_logbinop st g acl r iff negated
		    | (RNT(j,_,_)::_,_) when j = globalids.ddif -> raise BadParseException
		    | (RNT(j,_,_)::_,_) when j = globalids.ddotherwise -> raise BadParseException
		    | _ ->
			let (h,v) = extract_genterm_head_name gtl false [] in
			let frlr = ref [] in
			let ml =
			  List.map
			    (fun c ->
			      let (m,frl) = mizar_elab_and_test_rec st g c in
			      frlr := union_frees frl !frlr;
			      m)
			    acl
			in
			let prem = genpreap (PName(h)) (List.map (fun arg -> PTrm(arg)) ml) in (*** use the term version embedded as a preterm, to avoid needing to elaborate subterms multiple times ***)
			let (_,_,_,_,_,(_,u,_)) = st in
			u := false;
			let (mtm,frl) = mizar_elaborate_trm_force_stp st g prem Prop in
			let mtm = if negated then Ap(Neg,mtm) else mtm in
			let frl = union_frees frl !frlr in
			add_precomp g r (mtm,Stp(Prop),frl);
			(mtm,frl)
	      else if i = globalids.numi then
		begin
		  match cl with
		  | [RT(i,_);(RNT(_,r1,_) as tr1);RT(j,_)] when i = globalids.lparen && j = globalids.rparen ->
		      begin
			let n = mizar_elab_and_test_rec st g tr1 in
			r := !r1;
			n
		      end
		  | [RT(j,rj)] ->
		      let xj = string_of_identifier j in
		      let prem = PName(xj) in
		      let (_,_,_,_,_,(_,u,_)) = st in
		      u := false;
		      let (mtm,frl) = mizar_elaborate_trm_force_stp st g prem itp in
		      let gtp = stp_to_gentp st mtm itp in
		      add_precomp g r (mtm,gtp,frl);
		      (mtm,frl)
		  | _ -> raise BadParseException
		end
	      else if i = globalids.seti then
		begin
		  match cl with
		  | [RT(i,_);(RNT(_,r1,_) as tr1);RT(j,_)] when i = globalids.lparen && j = globalids.rparen ->
		      begin
			let n = mizar_elab_and_test_rec st g tr1 in
			r := !r1;
			n
		      end
		  | [RNT(i,_,_) as frch] when i = globalids.fraenkel ->
		      mizar_elab_and_test_rec st g frch
		  | [RNT(i,_,_) as tr1] when i = globalids.numi ->
		      mizar_elab_and_test_rec st g tr1
		  | [RNT(i,_,_);tr2] when i = globalids.ddthe ->
		      let (m,frl) = mizar_elab_and_test_sort st g tr2 in
		      let pl = extract_conjuncts m true in
		      let cn = Ap(Choice(itp),Lam(itp,m)) in
		      let nl = List.map (fun p -> subst p 0 cn) pl in
		      mizar_widen st nl mizwidenlimit;
		      let gtp = Mtp(nl) in
		      add_precomp g r (cn,gtp,frl);
		      (cn,frl)
		  | _ ->
		      let (tl,gtl,acl,oth) = extract_children cl [] [] [] [] in
		      if not (tl = []) then
			begin
			  Printf.printf "#+#++ Unexpected %d terminal children %s\n" (List.length tl) (string_of_rich_grammar_tree_type tr);
			  raise BadParseException
			end;
		      match (gtl,acl) with
		      | ([],[RNT(j,rj,_) as trj]) when i = j -> (*** intended as casting ***)
			  begin
			    let (n,frl) = mizar_elab_and_test_rec st g trj in
			    r := !rj;
			    (n,frl)
			  end
		      | _ ->
			  let (h,v) = extract_genterm_head_name gtl false [] in
			  let frlr = ref [] in
			  let ml =
			    List.map
			      (fun c ->
				let (m,frl) = mizar_elab_and_test_rec st g c in
				frlr := union_frees frl !frlr;
				m)
			      acl
			  in
			  let prem = genpreap (PName(h)) (List.map (fun arg -> PTrm(arg)) ml) in (*** use the term version embedded as a preterm, to avoid needing to elaborate subterms multiple times ***)
			  let (_,_,_,_,_,(_,u,_)) = st in
			  u := false;
			  let (mtm,frl) = mizar_elaborate_trm_force_stp st g prem itp in
			  let frl = union_frees frl !frlr in
			  let gtp = stp_to_gentp st mtm itp in
			  add_precomp g r (mtm,gtp,frl);
			  (mtm,frl)
		end
	      else if typish_id_p i then
		mizar_elab_and_test_sort st g tr
	      else
		raise (DontKnowException((string_of_identifier i) ^ " nonterminal cannot be at the top"))
	end
    | RT(i,r) -> (*** actually, I should never reach this point ***)
	begin
	  Printf.printf "#+#++ reached a terminal instead of generalized terminal %s\n" (string_of_identifier i);
	  raise (Failure("reached a terminal"))
	end
and mizar_elab_and_test_all r st g cl negated =
  if debug then
    begin
      Printf.printf "_all Input: %s\n" (string_of_rich_grammar_tree_type_l cl);
      Printf.printf "g:\n";
      List.iter (fun (y,n) -> Printf.printf "%s ~> %s\n" y (trm_str n)) g
    end;
  let (ta,eh,sh,sc,cc,(varh,u,varcnt)) = st in
  let (xl,vtp,suchth,body) = extract_univ_parts cl in
  if debug then
    begin
      Printf.printf "vtp: %s\n" (match vtp with Some(vtp) -> string_of_rich_grammar_tree_type vtp | None -> "None");
    end;
  let g2 = ref g in
  let zl = ref [] in
  let xxl = ref [] in
  List.iter
    (fun y ->
      let y2 = nxtfreshvar itp st in
      if debug then Printf.printf "#+#++ %s ~> %s\n" y (trm_str y2);
      xxl := (y,y2)::!xxl;
      g2 := (y,y2)::!g2;
      match y2 with Name(z,_) -> zl := (z,PIndTp)::!zl | _ -> ())
    xl;
  zl := List.rev !zl;
  begin (*** work out the types of the new vars first, before trying to elaborate the st or body ***)
    match vtp with
    | Some(vtp) ->
	List.iter
	  (fun (y,y2) ->
	    ignore (mizar_elab_and_test_is_sort st g vtp y y2))
	  !xxl; (*** reverse it, so that it starts with the radix and progressively adds adjectives; otherwise it isn't clear if the adjectives apply ***)
    | None ->
	()
  end;
  let (mbody,frlbody) = mizar_elab_and_test_rec st !g2 body in
  let mbodyd = ref mbody in
  List.iter (fun (x,x2) -> mbodyd := abstract_subterms 0 x2 (shift !mbodyd 0 1)) !xxl;
  let xxlf = List.rev !xxl in
  match xxlf with
  | [] -> raise BadParseException (*** disallow forall with no vars bound ***)
  | (x,x2)::xxr ->
      begin
	if tpof mbody = Prop then
	  begin
	    match suchth with
	    | Some(suchth) ->
		begin
		  let (msuchth,frlsuchth) = mizar_elab_and_test_rec st !g2 suchth in
		  if tpof msuchth = Prop then
		    begin
		      let msuchthd = ref msuchth in
		      List.iter (fun (x,x2) -> msuchthd := abstract_subterms 0 x2 (shift !msuchthd 0 1)) !xxl;
		      match vtp with
		      | Some(vtp) ->
			  let (mh,frltp) = mizar_elab_and_test_sort st g vtp in
			  let qtm = List.fold_left (fun q _ -> Ap(Forall(itp),Lam(itp,imp mh q))) (imp !msuchthd !mbodyd) !zl in
			  let frl = union_frees frltp (union_frees frlsuchth frlbody) in
			  let qtm = if negated then neg qtm else qtm in
			  add_precomp g r (qtm,Stp(Prop),frl);
			  (qtm,frl)
		      | None ->
			  let qtm = List.fold_left (fun q _ -> Ap(Forall(itp),Lam(itp,q))) (imp !msuchthd !mbodyd) !zl in
			  let frl = union_frees frlsuchth frlbody in
			  let qtm = if negated then neg qtm else qtm in
			  add_precomp g r (qtm,Stp(Prop),frl);
			  (qtm,frl)
		    end
		  else
		    raise (ExpectedTypeError(PTrm(msuchth),tpof msuchth,Prop))
		end
	    | None ->
		begin
		  match vtp with
		  | Some(vtp) ->
		      let (mh,frltp) = mizar_elab_and_test_sort st g vtp in
		      let qtm = List.fold_left (fun q _ -> Ap(Forall(itp),Lam(itp,imp mh q))) !mbodyd !zl in
		      let frl = union_frees frltp frlbody in
		      let qtm = if negated then neg qtm else qtm in
		      add_precomp g r (qtm,Stp(Prop),frl);
		      (qtm,frl)
		  | None ->
		      let qtm = List.fold_left (fun q _ -> Ap(Forall(itp),Lam(itp,q))) !mbodyd !zl in
		      let qtm = if negated then neg qtm else qtm in
		      add_precomp g r (qtm,Stp(Prop),frlbody);
		      (qtm,frlbody)
		end
	  end
	else
	  raise (ExpectedTypeError(PTrm(mbody),tpof mbody,Prop))
      end
and mizar_elab_and_test_ex r st g cl negated =
  let (ta,eh,sh,sc,cc,(varh,u,varcnt)) = st in
  let (xl,vtp,body) = extract_ex_parts cl in
  let g2 = ref g in
  let zl = ref [] in
  let xxl = ref [] in
  List.iter
    (fun y ->
      let y2 = nxtfreshvar itp st in
      if debug then Printf.printf "#+#++ %s ~> %s\n" y (trm_str y2);
      xxl := (y,y2)::!xxl;
      g2 := (y,y2)::!g2;
      match y2 with Name(z,_) -> zl := (z,PIndTp)::!zl | _ -> ())
    xl;
  zl := List.rev !zl;
  begin (*** work out the types of the new vars first, before trying to elaborate the st or body ***)
    match vtp with
    | Some(vtp) ->
	List.iter
	  (fun (y,y2) ->
	    ignore (mizar_elab_and_test_is_sort st g vtp y y2))
	  !xxl;
    | None ->
	()
  end;
  let (mbody,frlbody) = mizar_elab_and_test_rec st !g2 body in
  let mbodyd = ref mbody in
  List.iter (fun (x,x2) -> mbodyd := abstract_subterms 0 x2 (shift !mbodyd 0 1)) !xxl;
  let xxlf = List.rev !xxl in
  match xxlf with
  | [] -> raise BadParseException (*** disallow exists with no vars bound ***)
  | (x,x2)::xxr ->
      begin
	if tpof mbody = Prop then
	  begin
	    match vtp with
	    | Some(vtp) ->
		let (mh,frltp) = mizar_elab_and_test_sort st g vtp in
		let qtm = List.fold_left (fun q _ -> neg (Ap(Forall(itp),Lam(itp,neg (conj mh q))))) !mbodyd !zl in
		let frl = union_frees frltp frlbody in
		let qtm = if negated then neg qtm else qtm in
		add_precomp g r (qtm,Stp(Prop),frl);
		(qtm,frl)
	    | None ->
		let qtm = List.fold_left (fun q _ -> neg (Ap(Forall(itp),Lam(itp,neg q)))) !mbodyd !zl in
		let qtm = if negated then neg qtm else qtm in
		add_precomp g r (qtm,Stp(Prop),frlbody);
		(qtm,frlbody)
	  end
	else
	  raise (ExpectedTypeError(PTrm(mbody),tpof mbody,Prop))
      end
and mizar_elab_and_test_fraenkel r st g cl =
  if debug then
    begin
      Printf.printf "_fraenkel Input: %s\n" (string_of_rich_grammar_tree_type_l cl);
      Printf.printf "g:\n";
      List.iter (fun (y,n) -> Printf.printf "%s ~> %s\n" y (trm_str n)) g
    end;
  let (ta,eh,sh,sc,cc,(varh,u,varcnt)) = st in
  let (xlvtpl,body,cond) = extract_fraenkel_parts cl in
  let g2 = ref g in
  let zl = ref [] in
  let xxl = ref [] in
  let bvarct = ref 0 in
  let frlxxl = ref [] in
  List.iter
    (fun (xl,vtpo) ->
      match vtpo with
      | None ->
	  raise BadParseException (*** all bound vars in a Fraenkel must be bounded by a (setlike) Mizar type ***)
      | Some(vtp) ->
	  let (mh,frltp) = mizar_elab_and_test_sort st !g2 vtp in
	  frlxxl := union_frees frltp !frlxxl;
	  List.iter
	    (fun y ->
	      incr bvarct;
	      let y2 = nxtfreshvar itp st in
	      if debug then Printf.printf "#+#++ %s ~> %s\n" y (trm_str y2);
	      let boundtp = ref (Lam(itp,mh)) in
	      List.iter (fun (x,x2,_) -> boundtp := Lam(itp,abstract_subterms 0 x2 (shift !boundtp 0 1))) !xxl;
	      xxl := (y,y2,!boundtp)::!xxl;
	      g2 := (y,y2)::!g2;
	      match y2 with Name(z,_) -> zl := (z,PIndTp)::!zl | _ -> ())
	    xl)
    xlvtpl;
  zl := List.rev !zl;
  if !bvarct < 1 || !bvarct > 6 then raise (Failure("unhandled number of bound vars in a Fraenkel term"));
  let (mbody,frlbody) = mizar_elab_and_test_rec st !g2 body in
  if not (tpof mbody = itp) then raise (ExpectedTypeError(PTrm(mbody),tpof mbody,itp));
  let mbodyd = ref mbody in
  List.iter (fun (x,x2,_) -> mbodyd := Lam(itp,abstract_subterms 0 x2 (shift !mbodyd 0 1))) !xxl;
  let (mcond,frlcond) =
    match cond with
    | Some(cond) ->
	let (mcond,frlcond) = mizar_elab_and_test_rec st !g2 cond in
	if not (tpof mcond = Prop) then raise (ExpectedTypeError(PTrm(mcond),tpof mcond,Prop));
	let mcondd = ref mcond in
	List.iter (fun (x,x2,_) -> mcondd := Lam(itp,abstract_subterms 0 x2 (shift !mcondd 0 1))) !xxl;
	(!mcondd,frlcond)
    | None ->
	let mcondd = ref (Ap(Ap(Imp,False),False)) in
	List.iter (fun _ -> mcondd := Lam(itp,!mcondd)) !xxl;
	(!mcondd,[])
  in
  let xxlf = List.rev !xxl in
  let mainname = Printf.sprintf "replSep%d" !bvarct in
  let mainr = ref (Name(mainname,Hashtbl.find !mizar_name_tp mainname)) in
  List.iter
    (fun (x,x2,xbd) ->
      mainr := Ap(!mainr,xbd))
    xxlf;
  (Ap(Ap(!mainr,!mbodyd),mcond),union_frees !frlxxl (union_frees frlbody frlcond))
and mizar_elab_and_test_logbinop st g acl r f negated =
  match acl with
  | [tr1;tr2] ->
      let (n1,frl1) = mizar_elab_and_test_rec st g tr1 in
      if tpof n1 = Prop then
	begin
	  let (n2,frl2) = mizar_elab_and_test_rec st g tr2 in
	  if tpof n2 = Prop then
	    begin
	      let n = f n1 n2 in
	      let frl = union_frees frl1 frl2 in
	      let n = if negated then neg n else n in
	      add_precomp g r (n,Stp(Prop),frl);
	      (n,frl)
	    end
	  else
	    raise (ExpectedTypeError(PTrm(n2),tpof n2,Prop))
	end
      else
	raise (ExpectedTypeError(PTrm(n1),tpof n1,Prop))
  | _ -> raise BadParseException  
and mizar_elab_and_test_is_sort st g tr x x2 =
  lasttr := tr;
  let (_,_,sh,_,_,_) = st in
  if debug then Printf.printf "#+#++ is_sort (%s) calling elab %s\n" x (string_of_rich_grammar_tree_type tr);
  let (m,frl) = mizar_elab_and_test_sort st g tr in
  if tpof m = Prop then
    begin
      if debug then Printf.printf "#+#++ is_sort x2 %s m %s\n" (trm_str x2) (trm_str m);
      let n = subst m 0 x2 in
      let pl = extract_conjuncts n true in
      if debug then Printf.printf "#+#++ is_sort (%s; %s) %s calling widen\n" x (trm_str x2) (trm_str n);
      mizar_widen st pl mizwidenlimit;
      if debug then Printf.printf "#+#++ is_sort (%s; %s) %s back from widen\n" x (trm_str x2) (trm_str n);
      (n,frl)
    end
  else
    raise (ExpectedTypeError(PTrm(m),tpof m,Prop))
and mizar_elab_and_test_sort st g tr = (*** leave first arg as DB(0,itp) ***)
  lasttr := tr;
  if debug then Printf.printf "#+#++ sort calling elab %s\n" (string_of_rich_grammar_tree_type tr);
  let y = nxtfreshvar itp st in
  let (mtmy,frl) = mizar_elab_and_test_sort_2 st (("",y)::g) tr y in (*** y is a fake unnamed var, first arg of typing preds ***)
  if debug then Printf.printf "#+#++ mtmy %s\n" (trm_str mtmy);
  let mtm = abstract_subterms 0 y mtmy in
  if debug then Printf.printf "#+#++ y %s mtm %s\n" (trm_str y) (trm_str mtm);
  (mtm,frl)
and mizar_elab_and_test_sort_2 st g tr y = (*** Use y as first arg, abstracted in prev func ***)
  if debug then Printf.printf "#+#++ sort 2 %s calling elab %s\n" (trm_str y) (string_of_rich_grammar_tree_type tr);
  match tr with
  | RNT(i,r,cl) ->
      begin
	match lookup_precomp g r with
	| Some(m,gtp,frl) -> (m,frl)
	| None ->
	    let h = string_of_identifier i in
	    if String.length h > 3 && h.[0] = '$' && h.[1] = '#' && h.[2] = '~' then
	      begin
	(*       Printf.printf "#+#++ WARNING: %s node being replaced with Type node\n" h; *)
		mizar_elab_and_test_rec st g (RNT(globalids.typei,r,cl))
	      end
	    else if i = globalids.atlistofadjectives then
	      begin
		match cl with
		| [tr1] ->
		    let (n1,frl1) = mizar_elab_and_test_sort_force_adj_or_type st g tr1 y in (*** now explicitly assume those in n1 before returning ***)
		    let pl = extract_conjuncts n1 true in
		    if debug then Printf.printf "#+#++ is_sort (%s) %s calling widen\n" (trm_str y) (trm_str n1);
		    mizar_widen st pl mizwidenlimit;
		    if debug then Printf.printf "#+#++ is_sort (%s) %s back from widen\n" (trm_str y) (trm_str n1);
		    (n1,frl1)
		| [tr1;tr2] ->
		    let (n2,frl2) = mizar_elab_and_test_sort_2 st g tr2 y in (*** it is important to process them in this order so that the radix is considered first and then adjectives refining the type are added afterwards; example: odd nat needs to know nat before using odd ***)
		    let (n1,frl1) = mizar_elab_and_test_sort_2 st g tr1 y in (*** now explicitly assume those in n1 before returning ***)
		    let pl = extract_conjuncts n1 true in
		    if debug then Printf.printf "#+#++ is_sort (%s) %s calling widen\n" (trm_str y) (trm_str n1);
		    mizar_widen st pl mizwidenlimit;
		    if debug then Printf.printf "#+#++ is_sort (%s) %s back from widen\n" (trm_str y) (trm_str n1);
		    let n = if n2 = True then n1 else if n1 = True then n2 else Ap(Ap(And,n1),n2) in
		    (n,union_frees frl1 frl2)
		| _ ->
		    Printf.printf "#+#++ ERROR: Unexpected number of children (%d) for @ListOfAdjectives node: %s\n" (List.length cl) (string_of_rich_grammar_tree_type !lasttr);
		    raise BadParseException
	      end
	    else if i = globalids.adjective then
	      begin
		match cl with
		| [RNT(j,_,_);(RNT(j2,_,_) as tr1)] when j = globalids.ddnon && j2 = globalids.attribute ->
		    let (n1,frl1) = mizar_elab_and_test_sort_2 st g tr1 y in
		    (Ap(Neg,n1),frl1)
		| [RNT(j2,_,_) as tr1] when j2 = globalids.attribute ->
		    mizar_elab_and_test_sort_2 st g tr1 y
		| [RNT(_,_,cr)] -> (*** look underneath to get to the non ***)
		    mizar_elab_and_test_sort_2 st g (RNT(globalids.adjective,ref None,cr)) y
		| _ ->
		    Printf.printf "#+#++ ERROR: Unexpected adjective node: %s\n" (string_of_rich_grammar_tree_type_l cl);
		    raise BadParseException
	      end
	    else if i = globalids.attribute then
	      begin
		mizar_elab_and_test_sort_3 st g cl y
	      end
	    else if typish_id_p i then
	      begin
		let cladj = ref [] in
		let clnotadj = ref [] in
		List.iter
		  (fun trj ->
		    match trj with
		    | RNT(j,_,_) when j = globalids.atlistofadjectives || j = globalids.adjective || j = globalids.attribute ->
			cladj := trj::!cladj
		    | _ -> clnotadj := trj::!clnotadj)
		  cl;
		if !clnotadj = [] then
		  begin
		    let (n1,frl1) = mizar_elab_and_test_sort_4 st g !cladj y in
		    let pl = extract_conjuncts n1 true in
		    if debug then Printf.printf "#+#++ is_sort (%s) %s calling widen\n" (trm_str y) (trm_str n1);
		    mizar_widen st pl mizwidenlimit;
		    if debug then Printf.printf "#+#++ is_sort (%s) %s back from widen\n" (trm_str y) (trm_str n1);
		    (n1,frl1)
		  end
		else
		  begin
		    let (n2,frl2) = mizar_elab_and_test_sort_3 st g (List.rev !clnotadj) y in
		    let pl = extract_conjuncts n2 true in
		    if debug then Printf.printf "#+#++ is_sort (%s) %s calling widen\n" (trm_str y) (trm_str n2);
		    mizar_widen st pl mizwidenlimit;
		    if debug then Printf.printf "#+#++ is_sort (%s) %s back from widen\n" (trm_str y) (trm_str n2);
		    let (n1,frl1) = mizar_elab_and_test_sort_4 st g !cladj y in
		    let pl = extract_conjuncts n1 true in
		    if debug then Printf.printf "#+#++ is_sort (%s) %s calling widen\n" (trm_str y) (trm_str n1);
		    mizar_widen st pl mizwidenlimit;
		    if debug then Printf.printf "#+#++ is_sort (%s) %s back from widen\n" (trm_str y) (trm_str n1);
		    let n = if n2 = True then n1 else if n1 = True then n2 else Ap(Ap(And,n1),n2) in
		    (n,union_frees frl1 frl2)
		  end
	      end
	    else
	      begin
		Printf.printf "#+#++ tree expected to be a sort but is not a Type node %s\n" (string_of_rich_grammar_tree_type tr);
		raise BadParseException
	      end
      end
  | RT(i,r) ->
      begin
	Printf.printf "#+#++ reached a terminal instead of generalized terminal %s\n" (string_of_identifier i);
	raise (Failure("reached a terminal when trying to parse a sort"))
      end
and mizar_elab_and_test_sort_3 st g cl y =
  if debug then Printf.printf "#+#++ sort 3 %s calling elab %s\n" (trm_str y) (string_of_rich_grammar_tree_type_l cl);
  let (tl,gtl,acl,oth) = extract_children cl [] [] [] [] in
  if not (tl = []) then
    begin
      Printf.printf "#+#++ ERROR: Unexpected terminal children %d %s\n" (List.length tl) (string_of_rich_grammar_tree_type_l cl);
      raise BadParseException
    end;
  if not (List.length gtl = 1) then
    begin
      Printf.printf "#+#++ ERROR: (2) not = 1 but %d: %s ; the node: %s\n" (List.length gtl) (string_of_rich_grammar_tree_type_l gtl) (string_of_rich_grammar_tree_type_l cl);
      raise BadParseException
    end;
  let (h,v) = extract_genterm_head_name gtl false [] in
  let frlr = ref [] in
  let ml =
    List.map
      (fun c ->
	let (m,frl) = mizar_elab_and_test_rec st g c in
	frlr := union_frees frl !frlr;
	m)
      acl
  in
  let prem = genpreap (PName(h)) (PTrm(y)::(List.map (fun arg -> PTrm(arg)) ml)) in (*** use the term version embedded as a preterm, to avoid needing to elaborate subterms multiple times ***)
  let (_,_,_,_,_,(_,u,_)) = st in
  u := false;
  let (mtm,frl) = mizar_elaborate_trm_force_stp st g prem Prop in
  let frl = union_frees frl !frlr in
  (mtm,frl)
and mizar_elab_and_test_sort_4 st g cl y =
  if debug then Printf.printf "#+#++ sort 4 %s calling elab %s\n" (trm_str y) (string_of_rich_grammar_tree_type_l cl);
  match cl with
  | [] -> (True,[])
  | (tr::cr) ->
      let (n2,frl2) = mizar_elab_and_test_sort_2 st g tr y in
      let (n1,frl1) = mizar_elab_and_test_sort_4 st g cr y in
      let n = if n2 = True then n1 else if n1 = True then n2 else Ap(Ap(And,n1),n2) in
      (n,union_frees frl1 frl2)
and mizar_elab_and_test_sort_force_adj_or_type st g tr y =
  match tr with
  | RNT(i,_,_) when i = globalids.adjective -> mizar_elab_and_test_sort_2 st g tr y
  | _ -> mizar_elab_and_test_sort_2 st g (RNT(globalids.typei,ref None,[tr])) y

let mizar_elab_and_test_real tmout tr =
  let u = ref false in
  try (*** second try for the improbable case that Timeout is raised after an exception was called but before set_timer 0.0 executes ***)
    begin
      try
	if !raiseunknownname && unknown_name_occurs_p [] tr then (*** if there are unknown names used in the tree, then don't bother trying to elaborate/typecheck, just say DontKnow with 1.0 ***)
	  (DontKnow,1.0)
	else
	  let metab : (pretrm,trm * stp * ((string * (trm list)) list)) Hashtbl.t = Hashtbl.create 100 in
	  let melabh : (trm,trm) Hashtbl.t = Hashtbl.create 100 in
	  let mtph : (trm,trm) Hashtbl.t = Hashtbl.create 100 in
	  let mtpc : (trm,unit) Hashtbl.t = Hashtbl.create 100 in
	  let mccv : (trm,unit) Hashtbl.t = Hashtbl.create 100 in
	  let mccp : (trm,trm) Hashtbl.t = Hashtbl.create 100 in
	  let mccb : (trm,trm) Hashtbl.t = Hashtbl.create 100 in
	  let mccf : (trm,trm) Hashtbl.t = Hashtbl.create 100 in
	  let mcc = (mccv,mccp,mccb,mccf) in
	  let varh : (string,unit) Hashtbl.t = Hashtbl.create 100 in
	  let mizstate = (metab,melabh,mtph,mtpc,mcc,(varh,u,mizvarcnt)) in
	  set_timer tmout;
	  if debug then Printf.printf "========= fresh mizstate =========\n";
	  Hashtbl.clear main_assumed_sorts;
	  Hashtbl.clear main_queried_sorts;
	  assume_needed_sorts := 0;
	  let (_,frl) = mizar_elab_and_test_rec mizstate [] tr in
	  set_timer 0.0;
	  let num_assumed = count_assumed frl in
	  let num_queried = count_queried frl in
	  let p = if num_queried > 0 then 1.0 -. (float_of_int num_assumed /. float_of_int num_queried) else 1.0 in
	  Printf.printf "#+#++ elaboration success with %d; %d; p = %f\n" num_assumed num_queried p;
	  if p < 0.75 then (*** don't count it if it had to assume too many of the required sorting axioms ***)
	    (CheckingFailure,0.0)
	  else
	    (ElaborationSuccess,p)
      with
      | Timeout ->
	  begin
	    set_timer 0.0;
	    (MizarTimeout,1.0)
	  end
      | UnknownName(x) ->
	  begin
	    set_timer 0.0;
	    Printf.printf "#+#++ Balking due to unknown name %s\n" x;
	    (DontKnow,1.0)
	  end
      | ElaborationSimpleTypeError(mtm,b,a) -> (*** fundamental simple typing problem, definite rejection ***)
	  begin
	    set_timer 0.0;
	    Printf.printf "#+#++ ElaborationSimpleTypeError %s; got %s; expected %s\n" (trm_str mtm) (stp_str b) (stp_str a);
	    (CheckingFailure,0.0)
	  end
      | ExpectedTypeError(m1,b,a) -> (*** fundamental simple typing problem, definite rejection ***)
	  begin
	    set_timer 0.0;
	    Printf.printf "#+#++ ExpectedTypeError %s; got %s; expected %s\n" (pretrm_str m1) (stp_str b) (stp_str a);
	    (CheckingFailure,0.0)
	  end
      | GenericSyntaxError(x) -> (*** syntax error, possibly due to insufficient annotations, strong rejection ***)       
	  begin
	    set_timer 0.0;
	    Printf.printf "#+#++ GenericSyntaxError %s\n" x;
	    (CheckingFailure,0.0)
	  end
      | ElaborationFailure(x) -> (*** elaboration failure, which might just be due to the elaboration/type checking algorithm being too weak, possible rejection ***)
	  begin
	    set_timer 0.0;
	    Printf.printf "#+#++ ElaborationFailure %s%s\n" x (if !u then " but passing due to unknown vars" else "");
	    if !u then
	      (DontKnow,0.9)
	    else
	      (CheckingFailure,0.0)
	  end
      | BadParseException ->
	  begin
	    set_timer 0.0;
	    (BadParse,0.0)
	  end
      | DontKnowException(msg) ->
	  begin
	    set_timer 0.0;
	    (DontKnow,1.0)
	  end
      | e ->
	  begin
	    set_timer 0.0;
	    Printf.printf "#+#++ Exception %s\n" (Printexc.to_string e);
	    (DontKnow,1.0) (*** Actually, I don't know what happened here. Just let it pass. ***)
	  end
    end
  with Timeout -> (*** second catch of Timeout for the improbable case that Timeout is raised after an exception was called but before set_timer 0.0 executes ***)
    set_timer 0.0;
    (MizarTimeout,1.0)    

let meatcount = ref 0

let mizar_elab_and_test tmout tr =
  incr meatcount; Printf.printf "#+#++ mizar_elab_and_test call %d\n" !meatcount;
  if debug then Printf.printf "#+#++ top Input: %s\n" (string_of_rich_grammar_tree_type tr);
  let (mer,p) = mizar_elab_and_test_real tmout tr in
  if debug then Printf.printf "#+#++ top Output: %s\n" (string_of_rich_grammar_tree_type tr);
  if debug then Printf.printf "mer = %s and p = %f\n" (mer_string mer) p;
  (mer,p)

let rec mizar_elab_and_test_allsub tmout tr =
  match tr with
  | RNT(i,r,cl) ->
      if (i = globalids.booli || i = globalids.seti || i = globalids.typei) then
	begin
	  let (mer,p) = mizar_elab_and_test tmout tr in
	  Printf.printf "(ALLSUB \"%s\" \"%s\" %f)\n" (string_of_rich_grammar_tree_type tr) (mer_string mer) p
	end;
      List.iter (fun trj -> mizar_elab_and_test_allsub tmout trj) cl
  | _ -> ()
    
let mizar_tree_to_thf tr =
  let b = Buffer.create 1000 in
(*  mizar_tree_to_thf_rec b [] tr; (* I clearly don't have time to do this. Just test on the tree directly. - Chad, Sep 05 2017 *) *)
  Buffer.contents b
    
